from serpent.window_controller import WindowController

import applescript


class DarwinWindowController(WindowController):

    def __init__(self):
        pass

    def locate_window(self, name):
        return name

    def move_window(self, window_id, x, y):
        applescript.AppleScript(f'''
            tell application "System Events" to tell window 1 of process "{window_id}"
                set position to { {x}, {y} }
            end tell
        ''').run()

    def focus_window(self, window_id):
        applescript.AppleScript(f'''
            tell application "System Events" to tell process "{window_id}"
                set frontmost to true
            end tell
        ''').run()

    def is_window_focused(self, window_id):
        focused_window_id = applescript.AppleScript('''
            tell application "System Events"
                set focusedWindow to name of first application process whose frontmost is true
                return focusedWindow
            end tell
        ''').run()

        return focused_window_id == window_id

    def get_window_geometry(self, window_id):
        geometry = dict()

        window_geometry = applescript.AppleScript(f'''
            tell application "System Events" to tell process "{window_id}"
                return get size of window 1
            end tell
        ''').run()

        geometry["width"] = int(window_geometry[0])
        geometry["height"] = int(window_geometry[1] - 20)

        window_information = applescript.AppleScript(f'''
            tell application "System Events" to tell window 1 of process "{window_id}"
                return get position
            end tell
        ''').run()

        geometry["x_offset"] = int(window_information[0])
        geometry["y_offset"] = int(window_information[1] + 20)

        return geometry



import random
import uuid
import os
import shutil


def create_training_and_validation_sets(file_paths, validation_set_probability=0.1, seed=None):
    if seed is None:
        seed = generate_seed()

    random.seed(seed)
    
    if isinstance(file_paths, str):
        file_paths = [file_paths]

    clear_current_dataset()
        
    for file_path in file_paths:
        file_names = os.listdir(file_path)
        random.shuffle(file_names)

        os.makedirs(f"datasets/current/training/{file_path.split(os.sep)[-1]}".replace("/", os.sep))
        os.makedirs(f"datasets/current/validation/{file_path.split(os.sep)[-1]}".replace("/", os.sep))
    
        for file_name in file_names:
            set_label = "training"
    
            if random.random() <= validation_set_probability:
                set_label = "validation"
    
            shutil.copyfile(
                f"{file_path}/{file_name}".replace("/", os.sep),
                f"datasets/current/{set_label}/{file_path.split(os.sep)[-1]}/{file_name}".replace("/", os.sep)
            )

    return seed


def generate_seed():
    return str(uuid.uuid4())


def clear_current_dataset():
    try:
        shutil.rmtree("datasets/current".replace("/", os.sep))
    except FileNotFoundError:
        pass

    os.mkdir("datasets/current".replace("/", os.sep))



import yaml
import yaml.scanner

import offshoot

config = dict()
plugin_config = dict()

try:
    with open("config/config.yml", "r") as f:
        try:
            config = yaml.safe_load(f) or {}
        except yaml.scanner.ScannerError as e:
            raise Exception("Configuration file at 'config/config.yml' contains invalid YAML...")
        except Exception as e:
            print(type(e))
except FileNotFoundError as e:
    raise Exception("Configuration file not found at: 'config/config.yml'...")

try:
    with open(offshoot.config["file_paths"]["config"], "r") as f:
        try:
            plugin_config = yaml.safe_load(f) or {}
        except yaml.scanner.ScannerError as e:
            raise Exception(f"Configuration file at '{offshoot.config['file_paths']['config']}' contains invalid YAML...")
        except Exception as e:
            print(type(e))
except FileNotFoundError as e:
    raise Exception(f"Configuration file not found at: '{offshoot.config['file_paths']['config']}'...")

config = {**plugin_config, **config}



import numpy as np


def meshgrid_around_center_for_shape(shape):
    # Y
    if shape[0] % 2 == 0:
        half_y = shape[0] // 2
        y_offset = np.arange(-half_y + 1, half_y, 1)
        y = np.r_[y_offset[:half_y], [0], y_offset[half_y:]]
    else:
        y = np.arange(-(shape[0] // 2), shape[0] // 2 + 1, 1)
    # X
    if shape[1] % 2 == 0:
        half_x = shape[1] // 2
        x_offset = np.arange(-half_x + 1, half_x, 1)
        x = np.r_[x_offset[:half_x], [0], x_offset[half_x:]]
    else:
        x = np.arange(-(shape[1] // 2), shape[1] // 2 + 1, 1)

    return np.meshgrid(x, y)


def distances_to_center(shape):
    x, y = meshgrid_around_center_for_shape(shape)
    points = np.array([y, x])

    return np.linalg.norm(points, axis=0)


def angles_to_center(shape):
    x, y = meshgrid_around_center_for_shape(shape)
    return np.rad2deg(np.arctan2(x, y)).astype("int16")



from serpent.visual_debugger.visual_debugger import VisualDebugger

import numpy as np


class GameFrameBufferError(BaseException):
    pass


class GameFrameBuffer:

    def __init__(self, size=5):
        self.size = size
        self.frames = list()

        self.visual_debugger = VisualDebugger()

    @property
    def full(self):
        return len(self.frames) >= self.size

    @property
    def previous_game_frame(self):
        return self.frames[0] if len(self.frames) else None

    def add_game_frame(self, game_frame):
        if self.full:
            self.frames = [game_frame] + self.frames[:-1]
        else:
            self.frames = [game_frame] + self.frames

    def to_visual_debugger(self):
        for i, game_frame in enumerate(self.frames):
            self.visual_debugger.store_image_data(
                np.array(game_frame.frame * 255, dtype="uint8"),
                game_frame.frame.shape,
                f"frame_{i + 1}"
            )



import random


class EpsilonGreedyQPolicy:

    def __init__(self, initial_epsilon=1.0, final_epsilon=0.1, max_steps=1000000):
        self.initial_epsilon = initial_epsilon
        self.final_epsilon = final_epsilon

        self.epsilon = self.initial_epsilon

        self.max_steps = max_steps

    def use_random(self):
        return random.random() <= self.epsilon

    def erode(self, factor=1):
        if self.epsilon > self.final_epsilon:
            self.epsilon -= (self.initial_epsilon - self.final_epsilon) / (self.max_steps / factor)
        else:
            self.epsilon = self.final_epsilon



from serpent.input_controller import InputController


class GameAPI:
    instance = None

    def __init__(self, game=None):
        self.game = game

        if not self.game.is_launched:
            self.game.launch(dry_run=True)

        self.input_controller = InputController(game=game)

        self.__class__.instance = self



from serpent.input_controller import InputController
from serpent.machine_learning.reinforcement_learning.replay_memory import ReplayMemory
from serpent.machine_learning.reinforcement_learning.epsilon_greedy_q_policy import EpsilonGreedyQPolicy

from serpent.visual_debugger.visual_debugger import VisualDebugger

from keras.models import Model, Sequential
from keras.layers import Dense, Activation, Flatten, Convolution2D, MaxPooling2D, AveragePooling2D, Input, merge
from keras.optimizers import Adam, rmsprop

import numpy as np

import random
import itertools

from termcolor import cprint


class DQN:

    def __init__(
        self,
        input_shape=None,
        input_mapping=None,
        replay_memory_size=10000,
        batch_size=32,
        action_space=None,
        max_steps=1000000,
        observe_steps=None,
        initial_epsilon=1.0,
        final_epsilon=0.1,
        gamma=0.99,
        model_file_path=None,
        model_learning_rate=2.5e-4,
        override_epsilon=False
    ):
        self.type = "DQN"
        self.input_shape = input_shape
        self.replay_memory = ReplayMemory(memory_size=replay_memory_size)
        self.batch_size = batch_size
        self.action_space = action_space
        self.action_count = len(self.action_space.combinations)
        self.action_input_mapping = self._generate_action_space_combination_input_mapping(input_mapping)
        self.frame_stack = None
        self.max_steps = max_steps
        self.observe_steps = observe_steps or (0.1 * replay_memory_size)
        self.current_observe_step = 0
        self.current_step = 0
        self.initial_epsilon = initial_epsilon
        self.final_epsilon = final_epsilon
        self.previous_epsilon = initial_epsilon
        self.epsilon_greedy_q_policy = EpsilonGreedyQPolicy(
            initial_epsilon=self.initial_epsilon,
            final_epsilon=self.final_epsilon,
            max_steps=self.max_steps
        )
        self.gamma = gamma
        self.current_action = None
        self.current_action_index = None
        self.current_action_type = None
        self.first_run = True
        self.mode = "OBSERVE"

        self.model_learning_rate = model_learning_rate
        self.model = self._initialize_model()

        if model_file_path is not None:
            self.load_model_weights(model_file_path, override_epsilon)

        self.model_loss = 0

        self.visual_debugger = VisualDebugger()

    def enter_train_mode(self):
        if self.previous_epsilon is not None:
            self.epsilon_greedy_q_policy.epsilon = self.previous_epsilon
            self.previous_epsilon = None

        self.mode = "TRAIN"

    def enter_run_mode(self):
        self.previous_epsilon = self.epsilon_greedy_q_policy.epsilon
        self.epsilon_greedy_q_policy.epsilon = 0.01
        self.mode = "RUN"

    def next_step(self):
        if self.mode == "TRAIN":
            self.current_step += 1
        elif self.mode == "OBSERVE":
            self.current_observe_step += 1

        if self.mode == "OBSERVE" and self.current_observe_step >= self.observe_steps:
            self.mode = "TRAIN"

    def build_frame_stack(self, game_frame):
        frame_stack = np.stack((
            game_frame,
            game_frame,
            game_frame,
            game_frame
        ), axis=2)

        self.frame_stack = frame_stack.reshape((1,) + frame_stack.shape)

    def update_frame_stack(self, game_frame_buffer):
        game_frames = [game_frame.eighth_resolution_grayscale_frame for game_frame in game_frame_buffer.frames]
        frame_stack = np.stack(game_frames, axis=2)

        self.frame_stack = frame_stack.reshape((1,) + frame_stack.shape)

    def append_to_replay_memory(self, game_frame_buffer, reward, terminal=False):
        previous_frame_stack = self.frame_stack
        self.update_frame_stack(game_frame_buffer)

        observation = [
            previous_frame_stack,
            self.current_action_index,
            reward,
            self.frame_stack,
            terminal
        ]

        self.replay_memory.add(self.calculate_target_error(observation), observation)

    def calculate_target_error(self, observation):
        previous_target = self.model.predict(observation[0])[0][observation[1]]

        if observation[4]:
            target = observation[2]
        else:
            target = observation[2] + self.gamma * np.max(self.model.predict(observation[3]))

        return np.abs(target - previous_target)

    def pick_action(self, action_type=None):
        if action_type is None:
            self.compute_action_type()
        else:
            self.current_action_type = action_type

        qs = self.model.predict(self.frame_stack)

        if self.current_action_type == "RANDOM":
            self.current_action_index = random.randrange(self.action_count)
            self.maximum_future_rewards = None
        elif self.current_action_type == "PREDICTED":
            self.current_action_index = np.argmax(qs)
            self.maximum_future_rewards = qs

    def compute_action_type(self):
        use_random = self.epsilon_greedy_q_policy.use_random()
        self.current_action_type = "RANDOM" if use_random else "PREDICTED"

    def erode_epsilon(self, factor=1):
        if self.mode == "TRAIN":
            self.epsilon_greedy_q_policy.erode(factor=factor)

    def generate_mini_batch(self):
        if self.mode == "OBSERVE":
            return None

        return self.replay_memory.sample(self.batch_size)

    def train_on_mini_batch(self):
        mini_batch = self.generate_mini_batch()

        flashback_indices = random.sample(range(self.batch_size), 6)

        for i in range(0, len(mini_batch)):
            if i in flashback_indices:
                flashback_image = np.squeeze(mini_batch[i][1][3][:, :, :, 1])

                self.visual_debugger.store_image_data(
                    np.array(flashback_image * 255, dtype="uint8"),
                    flashback_image.shape,
                    f"flashback_{flashback_indices.index(i) + 1}"
                )

                del flashback_image

            previous_frame_stack = mini_batch[i][1][0]
            action_index = mini_batch[i][1][1]
            reward = mini_batch[i][1][2]
            frame_stack = mini_batch[i][1][3]
            terminal = mini_batch[i][1][4]

            target = self.model.predict(previous_frame_stack)
            previous_target = target[action_index]

            projected_future_rewards = self.model.predict(frame_stack)

            if terminal:
                target[action_index] = reward
            else:
                target[action_index] = reward + self.gamma * np.max(projected_future_rewards)

            error = np.abs(target[action_index] - previous_target)
            self.replay_memory.update(mini_batch[i][0], error)

            self.model.fit(previous_frame_stack, target, epochs=1, verbose=0)

    def generate_action(self):
        self.current_action = self.action_space.combinations[self.current_action_index]

    def get_action_for_index(self, action_index):
        return [action_input.upper() for action_input in self.action_input_mapping[self.action_space.combinations[action_index]]]

    def get_input_values(self):
        return self.action_input_mapping[self.current_action]

    def save_model_weights(self, file_path_prefix="datasets/model_", is_checkpoint=False):
        epsilon = self.epsilon_greedy_q_policy.epsilon

        if is_checkpoint:
            file_path = f"{file_path_prefix}_dqn_{self.current_step}_{epsilon}_.h5"
        else:
            file_path = f"{file_path_prefix}_dqn_{epsilon}_.h5"

        self.model.save_weights(file_path, overwrite=True)

    def load_model_weights(self, file_path, override_epsilon):
        self.model.load_weights(file_path)
        self.model.compile(loss="logcosh", optimizer=Adam(lr=self.model_learning_rate, clipvalue=10))

        *args, steps, epsilon, extension = file_path.split("_")
        self.current_step = int(steps)

        if override_epsilon:
            self.previous_epsilon = float(epsilon)
            self.epsilon_greedy_q_policy.epsilon = float(epsilon)

    def output_step_data(self):
        if self.mode in ["TRAIN", "OBSERVE"]:
            print(f"CURRENT MODE: {self.mode}")
        else:
            cprint(f"CURRENT MODE: {self.mode}", "grey", "on_yellow", attrs=["dark"])

        print(f"CURRENT STEP: {self.current_step}")

        if self.mode == "OBSERVE":
            print(f"CURRENT OBSERVE STEP: {self.current_observe_step}")
            print(f"OBSERVE STEPS: {self.observe_steps}")

        print(f"CURRENT EPSILON: {round(self.epsilon_greedy_q_policy.epsilon, 6)}")
        print(f"CURRENT RANDOM ACTION PROBABILITY: {round(self.epsilon_greedy_q_policy.epsilon * 100.0, 2)}%")
        print(f"LOSS: {self.model_loss}")

    def _initialize_model(self):
        input_layer = Input(shape=self.input_shape)

        tower_1 = Convolution2D(16, 1, 1, border_mode="same", activation="elu")(input_layer)
        tower_1 = Convolution2D(16, 3, 3, border_mode="same", activation="elu")(tower_1)

        tower_2 = Convolution2D(16, 1, 1, border_mode="same", activation="elu")(input_layer)
        tower_2 = Convolution2D(16, 3, 3, border_mode="same", activation="elu")(tower_2)
        tower_2 = Convolution2D(16, 3, 3, border_mode="same", activation="elu")(tower_2)

        tower_3 = MaxPooling2D((3, 3), strides=(1, 1), border_mode="same")(input_layer)
        tower_3 = Convolution2D(16, 1, 1, border_mode="same", activation="elu")(tower_3)

        merged_layer = merge([tower_1, tower_2, tower_3], mode="concat", concat_axis=1)

        output = AveragePooling2D((7, 7), strides=(8, 8))(merged_layer)
        output = Flatten()(output)
        output = Dense(self.action_count)(output)

        model = Model(input=input_layer, output=output)
        model.compile(rmsprop(lr=self.model_learning_rate, clipvalue=1), "mse")

        return model

    def _generate_action_space_combination_input_mapping(self, input_mapping):
        action_input_mapping = dict()

        for combination in self.action_space.combinations:
            combination_values = self.action_space.values_for_combination(combination)
            input_values = [input_mapping[combination_value] for combination_value in combination_values if combination_value is not None]

            action_input_mapping[combination] = list(itertools.chain.from_iterable(input_values))

        return action_input_mapping



import time

from datetime import datetime


class GameFrameLimiter:

    def __init__(self, fps=30):
        self.frame_time = 1 / fps
        self.started_at = None

    def start(self):
        self.started_at = datetime.utcnow()

    def stop_and_delay(self):
        duration = (datetime.utcnow() - self.started_at).microseconds / 1000000
        remaining_frame_time = self.frame_time - duration

        if remaining_frame_time > 0:
            time.sleep(remaining_frame_time)

    def benchmark(self):
        pass



import offshoot


class SerpentGamePlugin(offshoot.Plugin):
    name = "SerpentGamePlugin"
    version = "0.1.0"

    libraries = []

    files = [
        {"path": "serpent_game.py", "pluggable": "Game"}
    ]

    config = {
        "fps": 2
    }

    @classmethod
    def on_install(cls):
        print("\n\n%s was installed successfully!" % cls.__name__)

    @classmethod
    def on_uninstall(cls):
        print("\n\n%s was uninstalled successfully!" % cls.__name__)


if __name__ == "__main__":
    offshoot.executable_hook(SerpentGamePlugin)



from serpent.sprite_locator import SpriteLocator

import serpent.ocr

import pyautogui

import enum

import time


class MouseButton(enum.Enum):
    LEFT = 0
    MIDDLE = 1
    RIGHT = 2


class InputController:

    def __init__(self, game=None):
        self.game = game

        self.mouse_buttons = {
            MouseButton.LEFT.name: "left",
            MouseButton.MIDDLE.name: "middle",
            MouseButton.RIGHT.name: "right"
        }

        self.previous_key_collection_set = set()

        self.sprite_locator = SpriteLocator()

    @property
    def game_is_focused(self):
        return self.game.is_focused

    # Keyboard Actions
    def handle_keys(self, key_collection):
        key_collection_set = set(key_collection)

        keys_to_press = key_collection_set - self.previous_key_collection_set
        keys_to_release = self.previous_key_collection_set - key_collection_set

        for key in keys_to_press:
            self.press_key(key)

        for key in keys_to_release:
            self.release_key(key)

        self.previous_key_collection_set = key_collection_set

    def tap_keys(self, keys, duration=0.05):
        if self.game_is_focused:
            for key in keys:
                pyautogui.keyDown(key)

            time.sleep(duration)

            for key in keys:
                pyautogui.keyUp(key)

    def tap_key(self, key, duration=0.05):
        if self.game_is_focused:
            pyautogui.keyDown(key)

            time.sleep(duration)

            pyautogui.keyUp(key)

    def press_keys(self, keys):
        for key in keys:
            self.press_key(key)

    def press_key(self, key):
        if self.game_is_focused:
            pyautogui.keyDown(key)

    def release_keys(self):
        for key in self.previous_key_collection_set:
            self.release_key(key)

        self.previous_key_collection_set = set()

    def release_key(self, key):
        if self.game_is_focused:
            pyautogui.keyUp(key)

    def type_string(self, string, duration=0.05):
        if self.game_is_focused:
            pyautogui.typewrite(message=string, interval=duration)

    # Mouse Actions
    def click(self, button=MouseButton.LEFT, y=None, x=None, duration=0.25):
        if self.game_is_focused:
            pyautogui.moveTo(x, y, duration=duration)
            pyautogui.click(button=self.mouse_buttons.get(button.name, "left"))

    def click_screen_region(self, button=MouseButton.LEFT, screen_region=None):
        if self.game_is_focused:
            screen_region_coordinates = self.game.screen_regions.get(screen_region)

            x = (screen_region_coordinates[1] + screen_region_coordinates[3]) // 2
            x += self.game.window_geometry["x_offset"]

            y = (screen_region_coordinates[0] + screen_region_coordinates[2]) // 2
            y += self.game.window_geometry["y_offset"]

            self.click(button=button, y=y, x=x)

    def click_sprite(self, button=MouseButton.LEFT, sprite=None, game_frame=None):
        if self.game_is_focused:
            sprite_location = self.sprite_locator.locate(sprite=sprite, game_frame=game_frame)

            if sprite_location is None:
                return False

            x = (sprite_location[1] + sprite_location[3]) // 2
            x += self.game.window_geometry["x_offset"]

            y = (sprite_location[0] + sprite_location[2]) // 2
            y += self.game.window_geometry["y_offset"]

            self.click(button=button, y=y, x=x)

            return True

    def click_string(self, query_string, button=MouseButton.LEFT, game_frame=None, fuzziness=2, ocr_preset=None):
        if self.game_is_focused:
            string_location = serpent.ocr.locate_string(
                query_string,
                game_frame.frame,
                fuzziness=fuzziness,
                ocr_preset=ocr_preset,
                offset_x=game_frame.offset_x,
                offset_y=game_frame.offset_y
            )

            if string_location is not None:
                x = (string_location[1] + string_location[3]) // 2
                x += self.game.window_geometry["x_offset"]

                y = (string_location[0] + string_location[2]) // 2
                y += self.game.window_geometry["y_offset"]

                self.click(button=button, y=y, x=x)

                return True

            return False

    def drag(self, button=MouseButton.LEFT, x0=None, y0=None, x1=None, y1=None, duration=1):
        if self.game_is_focused:
            pyautogui.moveTo(x0, y0, duration=0.2)
            pyautogui.dragTo(x1, y1, button=button, duration=duration)

    def drag_screen_region_to_screen_region(self, button=MouseButton.LEFT, start_screen_region=None, end_screen_region=None, duration=1):
        if self.game_is_focused:
            start_screen_region_coordinates = self._extract_screen_region_coordinates(start_screen_region)
            end_screen_region_coordinates = self._extract_screen_region_coordinates(end_screen_region)

            self.drag(
                button=button,
                x0=start_screen_region_coordinates[0],
                y0=start_screen_region_coordinates[1],
                x1=end_screen_region_coordinates[0],
                y1=end_screen_region_coordinates[1],
                duration=duration
            )

    def scroll(self, y=None, x=None, clicks=1, direction="DOWN"):
        if self.game_is_focused:
            clicks = clicks * (1 if direction == "DOWN" else -1)
            pyautogui.scroll(clicks, x=x, y=y)

    def _extract_screen_region_coordinates(self, screen_region):
        screen_region_coordinates = self.game.screen_regions.get(screen_region)

        x = (screen_region_coordinates[1] + screen_region_coordinates[3]) // 2
        x += self.game.window_geometry["x_offset"]

        y = (screen_region_coordinates[0] + screen_region_coordinates[2]) // 2
        y += self.game.window_geometry["y_offset"]

        return x, y



import numpy as np
import random


class ReplayMemory:
    def __init__(self, memory_size=5000):
        self.tree = SumTree(memory_size)
        self.alpha = 0.6

    def add(self, error, observation):
        priority = self._get_priority(error)
        self.tree.add(priority, observation)

    def sample(self, quantity):
        samples = list()
        segment = self.tree.total() / quantity

        for i in range(quantity):
            s = random.uniform(a=segment * i, b=segment * (i + 1))
            index, priority, observation = self.tree.get(s)

            samples.append((index, observation))

        return samples

    def update(self, index, error):
        priority = self._get_priority(error)
        self.tree.update(index, priority)

    def _get_priority(self, error):
        return (error + 0.01) ** self.alpha


class SumTree:

    def __init__(self, size=5000):
        self.position = 0
        self.size = size
        self.tree = np.zeros(2 * size - 1)
        self.data = np.zeros(size, dtype=object)

    def total(self):
        return self.tree[0]

    def add(self, priority, data):
        index = self.position + self.size - 1

        self.data[self.position] = data
        self.update(index, priority)

        self.position += 1

        if self.position >= self.size:
            self.position = 0

    def update(self, index, priority):
        delta = priority - self.tree[index]

        self.tree[index] = priority
        self._propagate(index, delta)

    def get(self, s):
        index = self._retrieve(0, s)
        data_index = index - self.size + 1

        return index, self.tree[index], self.data[data_index]

    def _propagate(self, index, delta):
        parent = (index - 1) // 2

        self.tree[parent] += delta

        if parent != 0:
            self._propagate(parent, delta)

    def _retrieve(self, index, s):
        left = 2 * index + 1
        right = left + 1

        if left >= len(self.tree):
            return index

        if s <= self.tree[left]:
            return self._retrieve(left, s)
        else:
            return self._retrieve(right, s-self.tree[left])




class SpriteIdentifier:
    def __init__(self):
        self.sprites = dict()

    def identify(self, sprite, mode="SIGNATURE_COLORS", score_threshold=75):
        if mode == "SIGNATURE_COLORS":
            return self.identify_by_signature_colors(sprite, score_threshold=score_threshold)
        elif mode == "CONSTELLATION_OF_PIXELS":
            return self.identify_by_constellation_of_pixels(sprite, score_threshold=score_threshold)

    def identify_by_signature_colors(self, query_sprite, score_threshold=0):
        top_sprite_score = 0
        top_sprite_match = None

        for sprite_name, sprite in self.sprites.items():
            for sprite_signature_colors in sprite.signature_colors:
                for query_sprite_signature_colors in query_sprite.signature_colors:
                    signature_color_score = len(query_sprite_signature_colors & sprite_signature_colors)

                    if signature_color_score > top_sprite_score and round((signature_color_score / len(sprite_signature_colors)) * 100.0, 2) >= score_threshold:
                        top_sprite_score = signature_color_score
                        top_sprite_match = sprite_name

        return top_sprite_match or "UNKNOWN"

    def identify_by_constellation_of_pixels(self, query_sprite, score_threshold=0):
        top_sprite_score = 0
        top_sprite_match = None

        for sprite_name, sprite in self.sprites.items():
            for constellation_of_pixels in sprite.constellation_of_pixels:
                for i in range(query_sprite.image_data.shape[3]):
                    query_sprite_image = query_sprite.image_data[..., i]

                    constellation_of_pixels_score = 0

                    for pixel_coordinates, pixel_color in constellation_of_pixels.items():
                        if tuple(query_sprite_image[pixel_coordinates[0], pixel_coordinates[1], :]) == pixel_color:
                            constellation_of_pixels_score += 1

                    if constellation_of_pixels_score > top_sprite_score and round((constellation_of_pixels_score / len(constellation_of_pixels)) * 100.0, 2) > score_threshold:
                        top_sprite_score = constellation_of_pixels_score
                        top_sprite_match = sprite_name

        return top_sprite_match or "UNKNOWN"

    def register(self, sprite):
        self.sprites[sprite.name] = sprite



from serpent.machine_learning.context_classification.context_classifiers.svm_context_classifier import SVMContextClassifier
from serpent.machine_learning.context_classification.context_classifiers.cnn_inception_v3_context_classifier import CNNInceptionV3ContextClassifier




from serpent.sprite import Sprite


class SpriteLocator:

    def __init__(self, **kwargs):
        pass

    def locate(self, sprite=None, game_frame=None):
        constellation_of_pixel_images = sprite.generate_constellation_of_pixels_images()
        location = None

        for i in range(len(constellation_of_pixel_images)):
            constellation_of_pixels_item = list(sprite.constellation_of_pixels[i].items())[0]

            query_coordinates = constellation_of_pixels_item[0]
            query_rgb = constellation_of_pixels_item[1]

            rgb_coordinates = Sprite.locate_color(query_rgb, image=game_frame.frame)
            rgb_coordinates = list(map(lambda yx: (yx[0] - query_coordinates[0], yx[1] - query_coordinates[1]), rgb_coordinates))

            maximum_y = game_frame.frame.shape[0] - constellation_of_pixel_images[i].shape[0]
            maximum_x = game_frame.frame.shape[1] - constellation_of_pixel_images[i].shape[1]

            for y, x in rgb_coordinates:
                if y < 0 or x < 0 or y > maximum_y or x > maximum_x:
                    continue

                for yx, rgb in sprite.constellation_of_pixels[i].items():
                    if tuple(game_frame.frame[y + yx[0], x + yx[1], :]) != rgb:
                        break
                else:
                    location = (
                        y,
                        x,
                        y + constellation_of_pixel_images[i].shape[0],
                        x + constellation_of_pixel_images[i].shape[1]
                    )

        return location



from serpent.window_controller import WindowController

import win32gui


class Win32WindowController(WindowController):

    def __init__(self):
        pass

    def locate_window(self, name):
        return win32gui.FindWindow(None, name)

    def move_window(self, window_id, x, y):
        x0, y0, x1, y1 = win32gui.GetWindowRect(window_id)
        win32gui.MoveWindow(window_id, x, y, x1 - x0, y1 - y0, True)

    def focus_window(self, window_id):
        win32gui.SetForegroundWindow(window_id)

    def is_window_focused(self, window_id):
        focused_window_id = win32gui.GetForegroundWindow()
        return focused_window_id == window_id

    def get_window_geometry(self, window_id):
        geometry = dict()

        x, y, width, height = win32gui.GetClientRect(window_id)

        geometry["width"] = width
        geometry["height"] = height

        x0, y0, x1, y1 = win32gui.GetWindowRect(window_id)

        border_width = ((x1 - x0 - width) // 2)

        geometry["x_offset"] = x0 + border_width
        geometry["y_offset"] = y0 + (y1 - y0 - height - border_width)

        return geometry



from serpent.game_launcher import GameLauncher, GameLauncherException

import sys
import shlex
import subprocess
import webbrowser


class SteamGameLauncher(GameLauncher):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def launch(self, **kwargs):
        app_id = kwargs.get("app_id")
        app_args = kwargs.get("app_args")

        if app_id is None:
            raise GameLauncherException("An 'app_id' kwarg is required...")

        protocol_string = f"steam://run/{app_id}"

        if app_args is not None:
            args_list = [f"--{k}={v}" for k, v in app_args.items()]
            protocol_string += "/en/" + " ".join(args_list)

        if sys.platform in ["linux", "linux2"]:
            subprocess.call(shlex.split(f"xdg-open '{protocol_string}'"))
        elif sys.platform == "darwin":
            subprocess.call(shlex.split(f"open '{protocol_string}'"))
        elif sys.platform == "win32":
            webbrowser.open(f"{protocol_string}")



from serpent.machine_learning.context_classification.context_classifier import ContextClassifier

from serpent.visual_debugger.visual_debugger import VisualDebugger

import os
import pickle
import subprocess

import skimage.io
import skimage.transform
import skimage.color
import skimage.filters

import numpy as np

import sklearn.svm

from serpent.config import config


visual_debugger = VisualDebugger()


class SVMContextClassifier(ContextClassifier):

    def __init__(self, input_shape=None):
        super().__init__()
        self.input_shape = input_shape

    def train(self, epochs=3, preprocessing_func=None):
        data = list()
        targets = list()

        for root, dirs, files in os.walk("datasets/current/training"):
            if root != "datasets/current/training":
                current_label = root.split("/")[-1]

                for file in files:
                    if not file.endswith(".png"):
                        continue

                    file_path = f"{root}/{file}"
                    sample = skimage.io.imread(file_path)

                    targets.append(current_label)

                    preprocessed_sample = preprocessing_func(sample) if preprocessing_func is not None else self.preprocess_sample(sample)

                    visual_debugger.store_image_data(np.array(preprocessed_sample * 255, dtype="uint8"), preprocessed_sample.shape, bucket="svm_input")
                    visual_debugger.store_image_data(sample, sample.shape, bucket="original_frame")

                    if preprocessed_sample.dtype != "uint8":
                        preprocessed_sample = np.array(preprocessed_sample * 255, dtype="uint8")

                    data.append(preprocessed_sample.flatten())

        self.classifier = sklearn.svm.SVC(gamma=0.001, C=100)
        self.classifier.fit(data, targets)

    def validate(self, preprocessing_func=None, file_path=None):
        if self.classifier is None:
            self.load_classifier(file_path)

        sample_count = 0
        correct_sample_count = 0

        for root, dirs, files in os.walk("datasets/current/validation"):
            if root != "datasets/current/validation":
                current_label = root.split("/")[-1]

                for file in files:
                    if not file.endswith(".png"):
                        continue

                    sample_count += 1

                    file_path = f"{root}/{file}"
                    sample = skimage.io.imread(file_path)

                    predicted_class = self.predict(sample, preprocessing_func=preprocessing_func)

                    if predicted_class == current_label:
                        correct_sample_count += 1

                    subprocess.call(["clear"])
                    print(f"SVM Accuracy: {correct_sample_count / sample_count}")

    def predict(self, input_frame, preprocessing_func=None):
        preprocessed_frame = preprocessing_func(input_frame) if preprocessing_func is not None else self.preprocess_sample(input_frame)

        if preprocessed_frame.dtype != "uint8":
            preprocessed_frame = np.array(preprocessed_frame * 255, dtype="uint8")

        return self.classifier.predict([preprocessed_frame.flatten()])[0]

    def save_classifier(self, file_path):
        serialized_classifier = pickle.dumps(self.classifier)

        with open(file_path, "wb") as f:
            f.write(serialized_classifier)

    def load_classifier(self, file_path):
        with open(file_path, "rb") as f:
            self.classifier = pickle.loads(f.read())

    def preprocess_sample(self, sample):
        gray_sample = np.array(skimage.color.rgb2gray(sample) * 255, dtype="uint8")

        threshold = skimage.filters.threshold_local(gray_sample, 21)
        bw_sample = gray_sample > threshold

        return skimage.transform.resize(bw_sample, (60, 96), mode="reflect", order=0)
